/**********************************************************************
 * queue.c                                                 October 1999
 * Horms                                             horms@valinux.com.
 * VA Linux System                                 http://valinux.com./
 *
 * Data independent queue 
 *
 * Released under the terms of the GNU GPL
 *
 *
 * Useless Note: This code was origionally written as a queue of a
 *               fixed data type on some steps at St. Marys Public
 *               School, NSW, Australia in June 1997.
 *
 **********************************************************************/

#include "queue.h"

void free_queue_member(
  queue_member *mem, 
  void (*destroy_queue_t)(const void *)
);
 

/**********************************************************************
 * create_queue
 * Create a new, empty queue
 * pre: q: pounter to an unintialised queue
 * return: new, empty queue
 *         NULL on error
 **********************************************************************/

queue *create_queue(queue *q){
  if( (q=(queue *)malloc(sizeof(queue))) == NULL ){
    syslog(
      LOG_WARNING, 
      "create_queue: could not allocate memory: %s",
      strerror(errno)
    );
    return(NULL);
  }
  
  q->first=NULL;
  q->last=NULL;
  q->size=0;

  return q;
}


/**********************************************************************
 * push
 * push an element onto the queue
 * pre: q: queue
 *      value: elemnt to push onto the queue
 *      destroy_queue_t: pointer to a function to destroy queue elements
 * post: queue with element added
 *       NULL on error. On error, where possible the queue is destroyed.
 **********************************************************************/

queue *push(queue *q, queue_t value, void (*destroy_queue_t)(const void *)){
  queue_member *new;

  if( (new=(queue_member *)malloc(sizeof(queue_member))) == NULL){
    syslog(
      LOG_WARNING, 
      "push: could not allocate memory: %s",
      strerror(errno)
    );
    destroy_queue(q, (void (*)(const void *))destroy_queue_t);
    return(NULL);
  }

  /*Put in payload*/
  new->value=value;

  /*Update queue pointers*/
  new->next=q->first;
  new->prev=NULL;
  if(q->first==NULL){
    q->last=new;
  }
  else{
    q->first->prev = new;
  }
  q->first=new;

  /*Increment queue size*/
  q->size++;

  return q;
}


/**********************************************************************
 * pop 
 * Pop an element off the queue
 * pre: q: queue to pop the element off
 *      value: element removed from the queue is assigned to *value
 * post: queue with element removed
 *       NULL on error
 * Note: popping an emty queue results in NULL being returned
 **********************************************************************/

queue *pop(queue *q, queue_t *value){
  queue_member *old_last;

  if(q==NULL){
    syslog(
      LOG_WARNING, 
      "pop: poped an NULL queue",
      strerror(errno)
    );
    return(NULL);
  }

  if(q->last==NULL){
    syslog(
      LOG_WARNING, 
      "pop: poped an empty queue",
      strerror(errno)
    );
    return(NULL);
  }

  /*Grab payload*/
  *value = q->last->value;

  /*Fix pointers*/
  old_last = q->last;
  if ( (q->last=old_last->prev) == NULL ){
    q->first=NULL;
  }
  else{
    q->last->next=NULL;
  }
  
  /*Decrement queue size*/
  q->size--;

  free_queue_member(old_last, NULL);
  
  return q;
}


/**********************************************************************
 * destroy_queue
 * Destroy a queue, destroying each element present in the
 * queue first
 * pre: q: queue to destroy
 *      destroy_queue_t: pointer to a function to destroy a queue
 *      element
 * post: queue and all elements in the queue are destroyed
 *       No retrurn value
 **********************************************************************/

void destroy_queue(queue *q, void (*destroy_queue_t)(const void *)){
  queue_member *old_first;

  if(q==NULL){
    return;
  }

  while(q->first!=NULL){
    old_first = q->first;
    q->first = old_first->next;
    free_queue_member(old_first, (void (*)(const void *))destroy_queue_t);
  }

  free(q);
}


/**********************************************************************
 * queue_length
 * Return the number of elements in the queue
 * pre: q: queue to find the number of elements in
 * post: number of elements in the queue
 *       -1 on error
 **********************************************************************/

int queue_length(queue *q){
  if(q!=NULL){
    return q->size;
  }
  else{
    return -1;
  }
}
  


/**********************************************************************
 * free_queue_member
 * Free a queue member
 * pre: queue_member: element to free
 *      destroy_queue_t: ponter to function to destroy element
 * post: queue_member is freed
 *       no return value
 **********************************************************************/

void free_queue_member(
  queue_member *mem, 
  void (*destroy_queue_t)(const void *)
){
  if(mem==NULL){
    return;
  }

  if(mem->value!=NULL && destroy_queue_t!=NULL){
    destroy_queue_t(&(mem->value));
  }
  free(mem);
}
